/**************************************************************************
 *									  *
 *		Copyright ( C ) 1990, Silicon Graphics, Inc.		  *
 *									  *
 *  These coded instructions, statements, and computer programs  contain  *
 *  unpublished  proprietary  information of Silicon Graphics, Inc., and  *
 *  are protected by Federal copyright law.  They  may  not be disclosed  *
 *  to  third  parties  or copied or duplicated in any form, in whole or  *
 *  in part, without the prior written consent of Silicon Graphics, Inc.  *
 *									  *
 **************************************************************************/
/*
 * Internals for the MIDI streams mulitplexor
 */
#ident "$Revision: 1.10 $"

#ifndef	__MIDI_H_
#define	__MIDI_H_

#include <sys/ioctl.h>
#include <sys/time.h>

/* Primitives for putmsg/getmsg (M_PROTO) interface */
#define	MIDI_STAMPED	1		/* delta-stamped 1-3 byte messages */
#define MIDI_SYSEX	2		/* delta-stamped sysex message */
#define	MIDI_STAMPI	3		/* delta-stamped input messages */
#define	MIDI_SYSEXI	4		/* delta-stamped input sysex */

struct midistamps {
	long prim_type;			/* always MIDI_STAMPED */
	struct timeval tv[1];		/* 1 or more stamps */
};

struct midisysex {
	long prim_type;			/* always MIDI_SYSEX */
	long dev_id;			/* equivalent to midimessage.dev_id */
	struct timeval tv;
};

struct midistampi {
	long prim_type;			/* MIDI_STAMPI or MIDI_SYSEXI */
	long data_len;			/* # midimessages in data portion */
	struct timeval tv;		/* timestamp for data portion */
};

union midiprims {
	long prim_type;
	struct midisysex sysex;
	struct midistamps stamps;
	struct midistampi stampi;
};

struct midimessage {			/* for 1-3 byte messages, !sysex */
	u_int len:2;			/* msg len [0..3] */
	u_int dev_id:6;			/* getmsg src / putmsg dest */
	u_char bytes[3];		/* bytes[1]=status */
};

#ifdef _KERNEL
/* Names for standard MIDI status bytes */

enum MIDIstatus {
    /* For distinguishing channel number from status */
    MIDI_ChannelMask		= 0x0F,
    MIDI_StatusMask		= 0xF0,

    /* Channel voice messages */
    MIDI_ChannelVoice		= 0x80,
    MIDI_NoteOff		= 0x80,
    MIDI_NoteOn			= 0x90,
    MIDI_PolyKeyPressure	= 0xA0,
    MIDI_ControlChange		= 0xB0,
    MIDI_ChannelModeSelect	= 0xB0,
    MIDI_ProgramChange		= 0xC0,
    MIDI_ChannelPressure	= 0xD0,
    MIDI_PitchBendChange	= 0xE0,

    /* System messages */
    MIDI_SysEx			= 0xF0,		/* System Exclusive */

    /* System common */
    MIDI_SystemCommon		= 0xF1,
    MIDI_TimeCodeQuarterFrame	= 0xF1,
    MIDI_SongPositionPointer	= 0xF2,
    MIDI_SongSelect		= 0xF3,
    MIDI_Undefined1		= 0xF4,
    MIDI_Undefined2		= 0xF5,
    MIDI_TuneRequest		= 0xF6,
    MIDI_EOX			= 0xF7,		/* End of System Exclusive */

    /* System real time */
    MIDI_SystemRealTime		= 0xF8,
    MIDI_TimingClock		= 0xF8,
    MIDI_Undefined3		= 0xF9,
    MIDI_Start			= 0xFA,
    MIDI_Continue		= 0xFB,
    MIDI_Stop			= 0xFC,
    MIDI_Undefined4		= 0xFD,
    MIDI_ActiveSensing		= 0xFE,
    MIDI_SystemReset		= 0xFF,
    MIDI_Meta			= 0xFF		/* MIDI Files only */
};


/* Chain for "I_LINK"'ed lower streams */
struct midilink {
	queue_t *lwq;			/* top write queue of lower stream */
	queue_t *lrq;			/* top read queue of lower stream */
	mblk_t *sysbuf;                 /* buffer for sysex data */
	int l_index;			/* system ID from struct linkblk */
	unsigned int l_flags;		/* flags specific to this device */
	struct midimessage runin;	/* input running status */
	u_char rstatus_out;		/* output running status; 0=>clear */
	int pad:24;
};

/* struct midilink flag values */
#define	MIDI_SYSEXSTART		0x01	/* 1 = sysex started in inp stream */

struct midi {
	/* Streams stuff */
	queue_t *urq;			/* bottom read queue of upper stream */
	int deviceminor;		/* to find our way back */
	struct	msgb	*q_first;	/* first data block on "write queue" */
	struct	msgb	*q_last;	/* last data block on "write queue" */

	/* mux stuff */
	long lower;			/* default midilink for read/write */
	unsigned int flags;		/* see flag values below */

	/* Time keeping stuff */
	struct timeval lastwrite;	/* last time of write */
	struct timeval lastwerror;	/* error in lastwrite */
	struct timeval nextwrite;	/* next time of write */
	struct timeval lastread;	/* last time of read */
	struct timeval timeleft;        /* time remaining 'til queue empties */

	/* Time stats: cumwoverhd/ntimeouts=avg time taken by midi_timeout */
	struct timeval cumwoverhd;	/* cumulative write overhead */
	struct timeval unpausetime;	/* time of unpause */
	struct timeval lastempty;	/* time of last empty */
	int ntimeouts;			/* num calls to midi_timeout */
	int honor1;			/* num first calls to honorStamps */
	int honor2;			/* num second calls to honorStamps */
	int setalrms;			/* num calls to (fast_)timeout */
};


/* Stuff from master.d/midi */

/* Upper streams: */
extern int midi_cnt;			/* Actual number */
extern lock_t midi_lock;		/* Controls all access to the array */
extern struct midi *dev_midi[];		/* Actual active instantiations */

/* Lower streams: */
extern int midilink_cnt;		/* Actual number */
extern lock_t midilink_lock;		/* Controls all access to the array */
extern struct midilink *midi_links[];	/* Actual lower links */

#endif /* _KERNEL */

/* USER VISIBLE */

/* ioctl cmds */
#define MIOC_STAMPI	_IOW('M',1,int)		/* 1 = delaystamp input */
#define MIOC_TIMELEFT   _IOR('M',2,struct timeval) 
#define MIOC_CHMOD      _IOW('M',3,int)
#define MIOC_ACTIVE	_IOW('M',4,int)
#define MIOC_TIMERRST   _IOW('M',5,struct timeval) 

/* struct midi flags values */
#define	MIDI_STAMP_INPUT	0x01	/* 1 = delaystamp input */
#define MIDI_PAUSED		0x02	/* 1 = timer is paused */
#define MIDI_INSYSEX            0x04    /* 1 = parsing sysex, buffer */
#define MIDI_WRONLY             0x08    /* 1 = write only stream */
#define MIDI_RDONLY             0x10    /* 1 = read only stream */

#endif	/* __MIDI_H_ */
